/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "unnamed_internal.h"
#ifndef __MINGW32__
  #include <glob.h>
#endif

static void config_dump_node(state_t *state, const config_node_t *root, int indent)
{
  char *spaces = (char*)malloc((size_t)(indent+1));
  if(unlikely(spaces != NULL))
  {
    memset(spaces, (int)' ', (size_t)indent);
    spaces[indent] = '\0';
  }
  if(unlikely(root == NULL))
  {
    (void)lprintf(TRACE, "%s(null)", spaces);
  }
  else if(root->value.length != 0)
  {
    (void)lprintf(TRACE, "%s%s = %s", spaces, (char*)root->name.value, (char*)root->value.value);
  }
  else if(root->children != NULL)
  {
    int i;
    (void)lprintf(TRACE, "%s%s =", spaces, (char*)root->name.value);
    (void)lprintf(TRACE, "%s{", spaces);
    for(i = 0; i < root->child_count; i++) config_dump_node(state, &root->children[i], indent + 2);
    (void)lprintf(TRACE, "%s}", spaces);
  }
  free(spaces);
}

void config_dump(state_t *state)
{
  config_dump_node(state, &state->config, 0);
}

static int read_file(const char *filename, buffer_t *data)
{
  struct stat st;
  FILE *fp = fopen(filename, "rb");
  if(unlikely(fp == NULL)) return -1;
  memset(&st, 0, sizeof(st));
  if(unlikely(fstat(fileno(fp), &st) != 0))
  {
    (void)fclose(fp);
    return -1;
  }
  if(buffer_resize(data, (uint32_t)st.st_size) != 0)
  {
    buffer_free(data);
    (void)fclose(fp);
    return -1;
  }
  data->length = (uint32_t)st.st_size;
  data->value[data->length] = (uint8_t)'\0';
  if(unlikely(fread(data->value, (size_t)st.st_size, 1, fp) != 1))
  {
    buffer_free(data);
    (void)fclose(fp);
    return -1;
  }
  (void)fclose(fp);
  return 0;
}

static void config_free_all(config_node_t *root)
{
  if(likely(root != NULL))
  {
    int i;
    buffer_free(&root->name);
    buffer_free(&root->value);
    for(i = 0; i < root->child_count; i++) config_free_all(&root->children[i]);
    root->child_count = 0;
    if(likely(root->children != NULL))
    {
      free(root->children);
      root->children = NULL;
    }
  }
}

void config_free(state_t *state)
{
  config_free_all(&state->config);
}

static int next_token(buffer_t *data, buffer_t *token)
{
  uint32_t end;
  buffer_init(token);
  while(data->value[data->offset] != (uint8_t)'\0' && isspace((int)data->value[data->offset]) != 0) data->offset++;
  if(unlikely(data->value[data->offset] == (uint8_t)'\0')) return -1;
  while(data->value[data->offset] == (uint8_t)'/') /* start of a comment? */
  {
    if(data->value[data->offset + 1] == (uint8_t)'/')
    {
      while(data->value[data->offset] != (uint8_t)'\n' && data->value[data->offset] != (uint8_t)'\r' && data->value[data->offset] != (uint8_t)'\0') data->offset++;
    }
    else if(data->value[data->offset + 1] == (uint8_t)'*')
    {
      while(data->value[data->offset] != (uint8_t)'\0' && !(data->value[data->offset] == (uint8_t)'*' && data->value[data->offset + 1] == (uint8_t)'/')) data->offset++;
      if(data->value[data->offset] != (uint8_t)'\0') data->offset += 2;
    }
    else
    {
      break;
    }
    while(data->value[data->offset] != (uint8_t)'\0' && isspace((int)data->value[data->offset]) != 0) data->offset++;
  }
  if(data->value[data->offset] == (uint8_t)'=' || data->value[data->offset] == (uint8_t)'{' || data->value[data->offset] == (uint8_t)'}')
  {
    uint8_t c = data->value[data->offset];
    data->offset++;
    while(data->value[data->offset] != (uint8_t)'\0' && isspace((int)data->value[data->offset]) != 0) data->offset++;
    return buffer_put_raw(token, &c, 1);
  }

  end = data->offset;
  if(data->value[data->offset] == (uint8_t)'"' || data->value[data->offset] == (uint8_t)'\'')
  {
    end++;
    while(data->value[end] != data->value[data->offset] && data->value[end] != (uint8_t)'\0') end++;
    if(data->value[end] == (uint8_t)'\0') return -1;
    data->offset++;
    if(buffer_put_raw(token, data->value + data->offset, end - data->offset) != 0) return -1;
    data->offset = end + 1;
  }
  else
  {
    while(data->value[end] != (uint8_t)'\0' && isspace((int)data->value[end]) == 0 && data->value[end] != (uint8_t)'=' && data->value[end] != (uint8_t)'{' && data->value[end] != (uint8_t)'}') end++;
    if(buffer_put_raw(token, data->value + data->offset, end - data->offset) != 0) return -1;
    data->offset = end;
  }
  while(data->value[data->offset] != (uint8_t)'\0' && isspace((int)data->value[data->offset]) != 0) data->offset++;
  return 0;
}

/* todo: this is really bad ... */
static void child_pointer_cleanup(config_node_t *root)
{
  int i;
  for(i = 0; i < root->child_count; i++)
  {
    if(root->children[i].name.allocated == sizeof(root->children[i].name.small_value))
      root->children[i].name.value = root->children[i].name.small_value;
    if(root->children[i].value.allocated == sizeof(root->children[i].value.small_value))
      root->children[i].value.value = root->children[i].value.small_value;
  }
}

static config_node_t* add_child(config_node_t *root, int cleanup)
{
  if(unlikely(root->child_count % REALLOC_SIZE == 0))
  {
    config_node_t *temp;
    if(unlikely((temp = (config_node_t*)realloc(root->children, (size_t)(root->child_count + REALLOC_SIZE) * sizeof(root->children[0]))) == NULL)) return NULL;
    root->children = temp;
    if(cleanup != 0) child_pointer_cleanup(root);
  }
  memset(&root->children[root->child_count], 0, sizeof(root->children[0]));
  return &root->children[root->child_count++];
}

static int config_parse(buffer_t *data, config_node_t *root)
{
  int ret = -1;
  config_node_t *node;
  buffer_t name;
  buffer_init(&name);
  for(;;)
  {
    if(next_token(data, &name) != 0)
    {
      ret = 0;
      break;
    }

    if(strcmp((char*)name.value, "}") == 0)
    {
      ret = 0;
      break;
    }

    if(strcmp((char*)name.value, "include") == 0)
    {
      buffer_t filedata;
      buffer_init(&filedata);
      if(likely(next_token(data, &name) == 0))
      {
        #ifdef __MINGW32__
          if(likely(read_file((char*)name.value, &filedata) == 0))
          {
            if(config_parse(&filedata, root) != 0)
            {
              buffer_free(&filedata);
              break;
            }
          }
          buffer_free(&filedata);
        #else
          glob_t files;
          memset(&files, 0, sizeof(files));
          buffer_init(&filedata);
          if(glob((char*)name.value, 0, NULL, &files) == 0)
          {
            unsigned int i;
            for(i = 0; i < (unsigned int)files.gl_pathc; i++)
            {
              if(likely(read_file(files.gl_pathv[i], &filedata) == 0))
              {
                if(config_parse(&filedata, root) != 0)
                {
                  buffer_free(&filedata);
                  break;
                }
              }
              buffer_free(&filedata);
            }
            globfree(&files);
            if(i != (unsigned int)files.gl_pathc) break;
          }
        #endif
      }
      buffer_free(&name);
      continue;
    }

    if(data->value[data->offset] == (uint8_t)'=')
    {
      int i;
      data->offset++;
      for(i = 0; i < root->child_count; i++) if(buffer_cmp(&root->children[i].name, &name) == 0) break;
      if(i == root->child_count)
      {
        if(unlikely((node = add_child(root, 0)) == NULL)) break;
        if(buffer_copy(&node->name, &name) != 0) break;
      }
      else
      {
        node = &root->children[i];
      }
      if(likely(next_token(data, &node->value) != 0))
      {
        if(i == root->child_count)
        {
          buffer_free(&node->name);
          root->child_count--;
        }
        break;
      }
    }
    else /* not a name, assign an index */
    {
      if(unlikely((node = add_child(root, 0)) == NULL)) break;
      if(buffer_copy(&node->value, &name) != 0)
      {
        buffer_free(&node->name);
        root->child_count--;
        break;
      }
      buffer_free(&name);
      root->maxn++;
      if(unlikely(buffer_set(&node->name, "%d", root->maxn) != 0))
      {
        root->maxn--;
        buffer_free(&node->value);
        root->child_count--;
        break;
      }
    }
    if(strcmp((char*)node->value.value, "{") == 0)
    {
      buffer_free(&node->value);
      if(config_parse(data, node) != 0) break;
    }
  }
  buffer_free(&name);
  child_pointer_cleanup(root);
  return ret;
}

int config_read(const char *filename, config_node_t *root)
{
  buffer_t data;
  buffer_init(&data);
  if(unlikely(read_file(filename, &data) != 0)) return -1;
  config_free_all(root);
  if(buffer_set(&root->name, "root") != 0)
  {
    buffer_free(&data);
    return -1;
  }
  if(config_parse(&data, root) != 0)
  {
    buffer_free(&data);
    return -1;
  }
  buffer_free(&data);
  return 0;
}

static int name_equal(const buffer_t *a, const char *b, int b_length)
{
  if((int)a->length != b_length) return 0;
  if(memcmp(a->value, b, (size_t)b_length) == 0) return 1;
  return 0;
}

static config_node_t* config_find(config_node_t *root, const char *name)
{
  const char *search, *dot;
  config_node_t* ret = NULL;
  int i, length;
  dot = strchr(name, '.');
  length = (dot != NULL) ? (int)(dot - name) : (int)strlen(name);

  if(root->last > root->child_count - 1) root->last = 0;
  for(i = root->last; i < root->child_count; i++)
  {
    if(name_equal(&root->children[i].name, name, length) != 0)
    {
      if(dot == NULL)
      {
        root->last = i;
        return &root->children[i];
      }
      if((ret = config_find(&root->children[i], dot + 1)) != NULL) return ret;
    }
  }
  for(i = 0; i < root->last; i++)
  {
    if(name_equal(&root->children[i].name, name, length) != 0)
    {
      if(dot == NULL)
      {
        root->last = i;
        return &root->children[i];
      }
      if((ret = config_find(&root->children[i], dot + 1)) != NULL) return ret;
    }
  }

  /* find last dot and seach for that. */
  search = NULL;
  for(dot = strchr(name, '.'); dot != NULL; dot = strchr(dot + 1, '.')) search = dot + 1;
  if(search != NULL)
  {
    for(i = 0; i < root->last; i++)
    {
      if(strcmp((char*)root->children[i].name.value, search) == 0)
      {
        root->last = i;
        return &root->children[i];
      }
    }
    for(i = root->last; i < root->child_count; i++)
    {
      if(strcmp((char*)root->children[i].name.value, search) == 0)
      {
        root->last = i;
        return &root->children[i];
      }
    }
  }
  return NULL;
}

int config_set(state_t *state, config_node_t *root, const char *key, char *value)
{
  config_node_t* node;
  int i, length;
  const char *dot = strchr(key, '.');
  length = (dot != NULL) ? (int)(dot - key) : (int)strlen(key);

  if(unlikely(value == NULL))
  {
    config_free_node(state, key);
    return 0;
  }

  for(i = 0; i < root->child_count; i++)
  {
    if(name_equal(&root->children[i].name, key, length) != 0)
    {
      if(dot == NULL)
      {
        root->children[i].value.length = 0;
        return buffer_put_raw(&root->children[i].value, (uint8_t*)value, (uint32_t)strlen(value));
      }
      return config_set(state, &root->children[i], dot + 1, value);
    }
  }

  if((node = add_child(root, 1)) != NULL)
  {
    if(likely(buffer_put_raw(&node->name, (const uint8_t*)key, (uint32_t)length) == 0))
    {
      if(dot == NULL)
      {
        return buffer_put_raw(&node->value, (const uint8_t*)value, (uint32_t)strlen(value));
      }
      else
      {
        return config_set(state, node, dot + 1, value);
      }
    }
  }
  return -1;
}

void config_free_node(state_t *state, const char *name)
{
  config_node_t *node = (config_node_t*)config_find(&state->config, name);
  if(likely(node != NULL))
  {
    config_free_all(node);
  }
}

int config_maxn(state_t *state, const char *name)
{
  config_node_t *node = (config_node_t*)config_find(&state->config, name);
  if(likely(node != NULL)) return node->maxn;
  return 0;
}

const char* config_get_const(state_t *state, const char *name)
{
  const config_node_t *node = config_find(&state->config, name);
  if(likely(node != NULL))
  {
    if(likely(node->value.length > 0))
    {
      DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %s", name, node->value.value);
      return (const char*)node->value.value;
    }
  }
  DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = NULL", name);
  return NULL;
}

char* config_get(state_t *state, const char *name)
{
  const config_node_t *node = config_find(&state->config, name);
  if(likely(node != NULL))
  {
    if(likely(node->value.length > 0))
    {
      char *ret = (char*)malloc(node->value.length + 1);
      if(unlikely(ret == NULL))
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "malloc failed");
        return NULL;
      }
      memcpy(ret, node->value.value, node->value.length);
      ret[node->value.length] = '\0';
      DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %s", name, node->value.value);
      return ret;
    }
  }
  DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = NULL", name);
  return NULL;
}

char *config_get_lower(state_t *state, const char *name)
{
  const config_node_t *node = config_find(&state->config, name);
  if(likely(node != NULL))
  {
    if(likely(node->value.length > 0))
    {
      int i;
      char *ret = (char*)malloc(node->value.length + 1);
      if(unlikely(ret == NULL))
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "malloc failed");
        return NULL;
      }
      for(i = 0; i < (int)node->value.length; i++) ret[i] = (char)tolower((int)node->value.value[i]);
      ret[node->value.length] = '\0';
      DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %s", name, node->value.value);
      return ret;
    }
  }
  DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = NULL", name);
  return NULL;
}

int config_get_int(state_t *state, const char *name, int default_value)
{
  const config_node_t *node = config_find(&state->config, name);
  errno = 0;
  if(likely(node != NULL))
  {
    int ret = (int)strtol((char*)node->value.value, NULL, 0);
    if(unlikely(errno != 0)) ret = default_value;
    DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %d", name, ret);
    return ret;
  }
  DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %d (default)", name, default_value);
  return default_value;
}

double config_get_double(state_t *state, const char *name, double default_value)
{
  const config_node_t *node = config_find(&state->config, name);
  errno = 0;
  if(likely(node != NULL))
  {
    double ret = atof((char*)node->value.value);
    if(unlikely(errno != 0)) ret = default_value;
    DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %g", name, ret);
    return ret;
  }
  DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %g (default)", name, default_value);
  return default_value;
}

int config_get_ip(state_t *state, const char *name,  address_t *addr)
{
  const config_node_t *node = config_find(&state->config, name);
  if(likely(node != NULL))
  {
    char buffer[IPCHAR_SIZE];
    int ret = char2addr((char*)node->value.value, 0, addr);
    buffer[0] = '\0';
    DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %s", name, addr2char(buffer, addr));
    return ret;
  }
  memset(addr, 0, sizeof(*addr));
  DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = (null)", name);
  return -1;
}

int config_get_ip_range(state_t *state, const char *name, ip_range_t *range)
{
  const config_node_t *node = config_find(&state->config, name);
  if(likely(node != NULL))
  {
    char begin[IPCHAR_SIZE], end[IPCHAR_SIZE], *pos;
    begin[0] = '\0';
    end[0] = '\0';

    if((pos = strchr((char*)node->value.value, '-')) != NULL)
    {
      intptr_t length = pos - (char*)node->value.value;
      if(unlikely(length >= (intptr_t)sizeof(begin)))
      {
        memset(range, 0, sizeof(*range));
        ERROR_IF(state->trace) (void)lprintf(TRACE, "Unable to parse %s in %s", (char*)node->value.value, name);
        return -1;
      }
      memcpy(begin, node->value.value, (size_t)length);
      begin[length] = '\0';
      if(unlikely(char2addr(begin, 0, &range->begin) != 0))
      {
        memset(range, 0, sizeof(*range));
        ERROR_IF(state->trace) (void)lprintf(TRACE, "Unable to parse %s in %s", (char*)node->value.value, name);
        return -1;
      }
      if(unlikely(char2addr(pos + 1, 0, &range->end) != 0))
      {
        memset(range, 0, sizeof(*range));
        ERROR_IF(state->trace) (void)lprintf(TRACE, "Unable to parse %s in %s", (char*)node->value.value, name);
        return -1;
      }
      DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %s - %s", name, addr2char(begin, &range->begin), addr2char(end, &range->end));
      return 0;
    }

    if((pos = strchr((char*)node->value.value, '/')) != NULL)
    {
      int i, bits;
      intptr_t length = pos - (char*)node->value.value;
      if(unlikely(length >= (intptr_t)sizeof(begin)))
      {
        memset(range, 0, sizeof(*range));
        ERROR_IF(state->trace) (void)lprintf(TRACE, "Unable to parse %s in %s", node->value.value, name);
        return -1;
      }
      memcpy(begin, node->value.value, (size_t)length);
      begin[length] = '\0';
      pos++;

      if(unlikely(char2addr(begin, 0, &range->begin) != 0))
      {
        memset(range, 0, sizeof(*range));
        ERROR_IF(state->trace) (void)lprintf(TRACE, "Unable to parse %s in %s", node->value.value, name);
        return -1;
      }
      if(char2addr(pos, 0, &range->end) == 0)
      {
        if(range->begin.a.sockaddr.sa_family != range->end.a.sockaddr.sa_family)
        {
          memset(range, 0, sizeof(*range));
          ERROR_IF(state->trace) (void)lprintf(TRACE, "Address family mismatch %s in %s", (char*)node->value.value, name);
          return -1;

        }
        if(range->begin.a.sockaddr.sa_family == AF_INET)
        {
          range->end.a.sockaddr_in.sin_addr.s_addr = range->begin.a.sockaddr_in.sin_addr.s_addr | ~range->end.a.sockaddr_in.sin_addr.s_addr;
        }
        else if(range->begin.a.sockaddr.sa_family == AF_INET6)
        {
          for(i = 0; i < (int)sizeof(range->end.a.sockaddr_in6.sin6_addr.s6_addr); i++)
          {
            range->end.a.sockaddr_in6.sin6_addr.s6_addr[i] = (uint8_t)(range->begin.a.sockaddr_in6.sin6_addr.s6_addr[i] | ~range->end.a.sockaddr_in6.sin6_addr.s6_addr[i]);
          }
        }
        else
        {
          memset(range, 0, sizeof(*range));
          ERROR_IF(state->trace) (void)lprintf(TRACE, "Unknown family %u", range->begin.a.sockaddr.sa_family);
          return -1;
        }
        DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %s - %s", name, addr2char(begin, &range->begin), addr2char(end, &range->end));
        return 0;
      }
      errno = 0;
      bits = (int)strtol(pos, NULL, 0);
      if(unlikely(errno != 0 || bits < 0))
      {
        memset(range, 0, sizeof(*range));
        ERROR_IF(state->trace) (void)lprintf(TRACE, "Invalid bits %s in %s", (char*)node->value.value, name);
        return -1;
      }
      memcpy(&range->end, &range->begin, sizeof(range->end));
      if(range->begin.a.sockaddr.sa_family == AF_INET)
      {
        uint32_t e = (uint32_t)((0xffffffff >> (unsigned int)bits) | ntohl(range->begin.a.sockaddr_in.sin_addr.s_addr));
        range->end.a.sockaddr_in.sin_addr.s_addr = htonl(e);
      }
      else if(range->begin.a.sockaddr.sa_family == AF_INET6)
      {
        bits = 128 - bits;
        if(unlikely(bits < 0))
        {
          memset(range, 0, sizeof(*range));
          ERROR_IF(state->trace) (void)lprintf(TRACE, "Invalid bits %d in %s", bits, name);
          return -1;
        }
        for(i = 15; bits > 8; i--)
        {
          range->begin.a.sockaddr_in6.sin6_addr.s6_addr[i] = 0;
          range->end.a.sockaddr_in6.sin6_addr.s6_addr[i] = 0xff;
          bits -= 8;
        }
        range->begin.a.sockaddr_in6.sin6_addr.s6_addr[i] &= (unsigned char)(0xff ^ ((2U << (unsigned int)bits) - 1));
        range->end.a.sockaddr_in6.sin6_addr.s6_addr[i] |= (unsigned char)((2U << (unsigned int)bits) - 1);
      }
      else
      {
        memset(range, 0, sizeof(*range));
        ERROR_IF(state->trace) (void)lprintf(TRACE, "Unknown family %u", range->begin.a.sockaddr.sa_family);
        return -1;
      }

      DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %s - %s", name, addr2char(begin, &range->begin), addr2char(end, &range->end));
      return 0;
    }

    if(unlikely(char2addr((char*)node->value.value, 0, &range->begin) != 0))
    {
      memset(range, 0, sizeof(*range));
      ERROR_IF(state->trace) (void)lprintf(TRACE, "Unable to parse %s in %s", (char*)node->value.value, name);
      return -1;
    }
    memcpy(&range->end, &range->begin, sizeof(range->end));
    DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = %s - %s", name, addr2char(begin, &range->begin), addr2char(end, &range->end));
    return 0;
  }
  memset(range, 0, sizeof(*range));
  DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s = (null)", name);
  return -1;
}
