/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "unnamed.h"

typedef struct
{
  char *pattern, *label;
  uint32_t ttl;
  int stop;
} in_label_t;

typedef struct
{
  char *pattern;
  address_t address;
  uint32_t ttl;
  int stop;
} in_address_t;

typedef struct
{
  char *pattern, *exchange;
  uint32_t ttl;
  uint16_t preference;
  int stop;
} in_mx_t;

typedef struct
{
  char *pattern;
  char *mname, *rname;
  uint32_t serial, refresh, retry, expire, minimum;
  uint32_t ttl;
  int stop;
} in_soa_t;

typedef struct
{
  char *pattern;
  buffer_t txt;
  uint32_t ttl;
  int stop;
} in_txt_t;

typedef struct
{
  in_address_t *local_in_a;
  in_address_t *local_in_aaaa;
  in_label_t *local_in_cname;
  in_mx_t *local_in_mx;
  in_label_t *local_in_ns;
  in_label_t *local_in_ptr;
  in_soa_t *local_in_soa;
  in_txt_t *local_in_spf;
  in_txt_t *local_in_txt;

  int local_in_a_count, local_in_aaaa_count, local_in_cname_count, local_in_mx_count, local_in_ns_count, local_in_ptr_count, local_in_soa_count, local_in_spf_count, local_in_txt_count;
  int trace;
} module_data_t;

static void term(int module_index, state_t *state)
{
  int i;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  if(module->local_in_a != NULL)
  {
    for(i = 0; i < module->local_in_a_count; i++) free((void*)module->local_in_a[i].pattern);
    free(module->local_in_a);
  }
  if(module->local_in_aaaa != NULL)
  {
    for(i = 0; i < module->local_in_aaaa_count; i++) free((void*)module->local_in_aaaa[i].pattern);
    free(module->local_in_aaaa);
  }
  if(module->local_in_cname != NULL)
  {
    for(i = 0; i < module->local_in_cname_count; i++)
    {
      free(module->local_in_cname[i].pattern);
      free(module->local_in_cname[i].label);
    }
    free(module->local_in_cname);
  }
  if(module->local_in_mx != NULL)
  {
    for(i = 0; i < module->local_in_mx_count; i++)
    {
      free(module->local_in_mx[i].pattern);
      free(module->local_in_mx[i].exchange);
    }
    free(module->local_in_mx);
  }
  if(module->local_in_ns != NULL)
  {
    for(i = 0; i < module->local_in_ns_count; i++)
    {
      free(module->local_in_ns[i].pattern);
      free(module->local_in_ns[i].label);
    }
    free(module->local_in_ns);
  }
  if(module->local_in_ptr != NULL)
  {
    for(i = 0; i < module->local_in_ptr_count; i++)
    {
      free(module->local_in_ptr[i].pattern);
      free(module->local_in_ptr[i].label);
    }
    free(module->local_in_ptr);
  }
  if(module->local_in_soa != NULL)
  {
    for(i = 0; i < module->local_in_soa_count; i++)
    {
      free(module->local_in_soa[i].pattern);
      free(module->local_in_soa[i].mname);
      free(module->local_in_soa[i].rname);
    }
    free(module->local_in_soa);
  }
  if(module->local_in_spf != NULL)
  {
    for(i = 0; i < module->local_in_spf_count; i++)
    {
      free(module->local_in_spf[i].pattern);
      state->buffer_free(&module->local_in_spf[i].txt);
    }
    free(module->local_in_spf);
  }
  if(module->local_in_txt != NULL)
  {
    for(i = 0; i < module->local_in_txt_count; i++)
    {
      free(module->local_in_txt[i].pattern);
      state->buffer_free(&module->local_in_txt[i].txt);
    }
    free(module->local_in_txt);
  }
  free(module);
  state->module_data[module_index] = NULL;
}

static int init(int module_index, state_t *state)
{
  int count, i;
  buffer_t buffer;
  module_data_t *module = (module_data_t*)calloc(1, sizeof(*module));
  if(module == NULL) return -1;
  state->module_data[module_index] = module;

  #define CALLOC_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "calloc failed"); term(module_index, state); state->buffer_free(&buffer); return -1; }
  #define BUFFER_SET_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "buffer_set failed"); term(module_index, state); state->buffer_free(&buffer); return -1; }
  #define BUFFER_PUT_RAW_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "buffer_put_raw failed"); term(module_index, state); state->buffer_free(&buffer); return -1; }
  #define CONFIG_GET_IP_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "config_get_ip failed on %s", (char*)buffer.value); term(module_index, state); state->buffer_free(&buffer); return -1; }
  #define CONFIG_GET_LOWER_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "config_get_lower failed on %s", (char*)buffer.value); term(module_index, state); state->buffer_free(&buffer); return -1; }
  #define CONFIG_GET_CONST_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "config_get_const failed on %s", (char*)buffer.value); term(module_index, state); state->buffer_free(&buffer); return -1; }

  module->trace = state->config_get_int(state, "local.trace", TRACE_LEVEL_ERROR);

  buffer_init(&buffer);

  count = state->config_maxn(state, "local.a");
  if((module->local_in_a = (in_address_t*)calloc((size_t)count, sizeof(module->local_in_a[0]))) == NULL) CALLOC_FAILED;

  for(module->local_in_a_count = 0; module->local_in_a_count < count; module->local_in_a_count++)
  {
    memset(&module->local_in_a[module->local_in_a_count], 0, sizeof(module->local_in_a[0]));

    if(state->buffer_set(&buffer, "local.a.%d.ttl", module->local_in_a_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_a[module->local_in_a_count].ttl = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);

    if(state->buffer_set(&buffer, "local.a.%d.stop", module->local_in_a_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_a[module->local_in_a_count].stop = state->config_get_int(state, (char*)buffer.value, 0);

    if(state->buffer_set(&buffer, "local.a.%d.address", module->local_in_a_count + 1) != 0) BUFFER_SET_FAILED;
    if(state->config_get_ip(state, (char*)buffer.value, &module->local_in_a[module->local_in_a_count].address) != 0) CONFIG_GET_IP_FAILED;
    if(state->buffer_set(&buffer, "local.a.%d.pattern", module->local_in_a_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_a[module->local_in_a_count].pattern = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
  }

  count = state->config_maxn(state, "local.aaaa");
  if((module->local_in_aaaa = (in_address_t*)calloc((size_t)count, sizeof(module->local_in_aaaa[0]))) == NULL) CALLOC_FAILED;
  for(module->local_in_aaaa_count = 0; module->local_in_aaaa_count < count; module->local_in_aaaa_count++)
  {
    memset(&module->local_in_aaaa[module->local_in_aaaa_count], 0, sizeof(module->local_in_aaaa[0]));

    if(state->buffer_set(&buffer, "local.aaaa.%d.ttl", module->local_in_aaaa_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_aaaa[module->local_in_aaaa_count].ttl = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);

    if(state->buffer_set(&buffer, "local.aaaa.%d.stop", module->local_in_aaaa_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_aaaa[module->local_in_aaaa_count].stop = state->config_get_int(state, (char*)buffer.value, 0);

    if(state->buffer_set(&buffer, "local.aaaa.%d.address", module->local_in_aaaa_count + 1) != 0) BUFFER_SET_FAILED;
    if(state->config_get_ip(state, (char*)buffer.value,  &module->local_in_aaaa[module->local_in_aaaa_count].address) != 0) CONFIG_GET_IP_FAILED;
    if(state->buffer_set(&buffer, "local.aaaa.%d.pattern", module->local_in_aaaa_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_aaaa[module->local_in_aaaa_count].pattern = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
  }

  count = state->config_maxn(state, "local.cname");
  if((module->local_in_cname = (in_label_t*)calloc((size_t)count, sizeof(module->local_in_cname[0]))) == NULL) CALLOC_FAILED;
  for(module->local_in_cname_count = 0; module->local_in_cname_count < count; module->local_in_cname_count++)
  {
    memset(&module->local_in_cname[module->local_in_cname_count], 0, sizeof(module->local_in_cname[0]));

    if(state->buffer_set(&buffer, "local.cname.%d.ttl", module->local_in_cname_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_cname[module->local_in_cname_count].ttl = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);

    if(state->buffer_set(&buffer, "local.cname.%d.stop", module->local_in_cname_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_cname[module->local_in_cname_count].stop = state->config_get_int(state, (char*)buffer.value, 0);

    if(state->buffer_set(&buffer, "local.cname.%d.pattern", module->local_in_cname_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_cname[module->local_in_cname_count].pattern = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
    if(state->buffer_set(&buffer, "local.cname.%d.cname", module->local_in_cname_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_cname[module->local_in_cname_count].label = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
  }

  count = state->config_maxn(state, "local.mx");
  if((module->local_in_mx = (in_mx_t*)calloc((size_t)count, sizeof(module->local_in_mx[0]))) == NULL) CALLOC_FAILED;
  for(module->local_in_mx_count = 0; module->local_in_mx_count < count; module->local_in_mx_count++)
  {
    memset(&module->local_in_mx[module->local_in_mx_count], 0, sizeof(module->local_in_mx[0]));

    if(state->buffer_set(&buffer, "local.mx.%d.ttl", module->local_in_mx_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_mx[module->local_in_mx_count].ttl = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);

    if(state->buffer_set(&buffer, "local.mx.%d.stop", module->local_in_mx_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_mx[module->local_in_mx_count].stop = state->config_get_int(state, (char*)buffer.value, 0);

    if(state->buffer_set(&buffer, "local.mx.%d.preference", module->local_in_mx_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_mx[module->local_in_mx_count].preference = (uint16_t)state->config_get_int(state, (char*)buffer.value, 10);
    if(state->buffer_set(&buffer, "local.mx.%d.pattern", module->local_in_mx_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_mx[module->local_in_mx_count].pattern = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
    if(state->buffer_set(&buffer, "local.mx.%d.exchange", module->local_in_mx_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_mx[module->local_in_mx_count].exchange = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
  }

  count = state->config_maxn(state, "local.ns");
  if((module->local_in_ns = (in_label_t*)calloc((size_t)count, sizeof(module->local_in_ns[0]))) == NULL) CALLOC_FAILED;
  for(module->local_in_ns_count = 0; module->local_in_ns_count < count; module->local_in_ns_count++)
  {
    memset(&module->local_in_ns[module->local_in_ns_count], 0, sizeof(module->local_in_ns[0]));

    if(state->buffer_set(&buffer, "local.ns.%d.ttl", module->local_in_ns_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_ns[module->local_in_ns_count].ttl = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);

    if(state->buffer_set(&buffer, "local.ns.%d.stop", module->local_in_ns_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_ns[module->local_in_ns_count].stop = state->config_get_int(state, (char*)buffer.value, 0);

    if(state->buffer_set(&buffer, "local.ns.%d.pattern", module->local_in_ns_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_ns[module->local_in_ns_count].pattern = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
    if(state->buffer_set(&buffer, "local.ns.%d.nsdname", module->local_in_ns_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_ns[module->local_in_ns_count].label = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
  }

  count = state->config_maxn(state, "local.ptr");
  if((module->local_in_ptr = (in_label_t*)calloc((size_t)count, sizeof(module->local_in_ptr[0]))) == NULL) CALLOC_FAILED;
  for(module->local_in_ptr_count = 0; module->local_in_ptr_count < count; module->local_in_ptr_count++)
  {
    memset(&module->local_in_ptr[module->local_in_ptr_count], 0, sizeof(module->local_in_ptr[0]));

    if(state->buffer_set(&buffer, "local.ptr.%d.ttl", module->local_in_ptr_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_ptr[module->local_in_ptr_count].ttl = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);

    if(state->buffer_set(&buffer, "local.ptr.%d.stop", module->local_in_ptr_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_ptr[module->local_in_ptr_count].stop = state->config_get_int(state, (char*)buffer.value, 0);

    if(state->buffer_set(&buffer, "local.ptr.%d.pattern", module->local_in_ptr_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_ptr[module->local_in_ptr_count].pattern = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
    if(state->buffer_set(&buffer, "local.ptr.%d.ptrdname", module->local_in_ptr_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_ptr[module->local_in_ptr_count].label = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
  }

  count = state->config_maxn(state, "local.soa");
  if((module->local_in_soa = (in_soa_t*)calloc((size_t)count, sizeof(module->local_in_soa[0]))) == NULL) CALLOC_FAILED;
  for(module->local_in_soa_count = 0; module->local_in_soa_count < count; module->local_in_soa_count++)
  {
    memset(&module->local_in_soa[module->local_in_soa_count], 0, sizeof(module->local_in_soa[0]));

    if(state->buffer_set(&buffer, "local.soa.%d.ttl", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_soa[module->local_in_soa_count].ttl = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);

    if(state->buffer_set(&buffer, "local.soa.%d.stop", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_soa[module->local_in_soa_count].stop = state->config_get_int(state, (char*)buffer.value, 0);

    if(state->buffer_set(&buffer, "local.soa.%d.serial", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_soa[module->local_in_soa_count].serial = (uint32_t)state->config_get_int(state, (char*)buffer.value, (int)state->current_serial());
    if(state->buffer_set(&buffer, "local.soa.%d.refresh", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_soa[module->local_in_soa_count].refresh = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);
    if(state->buffer_set(&buffer, "local.soa.%d.retry", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_soa[module->local_in_soa_count].retry = (uint32_t)state->config_get_int(state, (char*)buffer.value, (int)module->local_in_soa[module->local_in_soa_count].refresh);
    if(state->buffer_set(&buffer, "local.soa.%d.expire", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_soa[module->local_in_soa_count].expire = (uint32_t)state->config_get_int(state, (char*)buffer.value, (int)module->local_in_soa[module->local_in_soa_count].refresh);
    if(state->buffer_set(&buffer, "local.soa.%d.minimum", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_soa[module->local_in_soa_count].minimum = (uint32_t)state->config_get_int(state, (char*)buffer.value, (int)module->local_in_soa[module->local_in_soa_count].refresh);
    if(state->buffer_set(&buffer, "local.soa.%d.pattern", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_soa[module->local_in_soa_count].pattern = state->config_get_lower(state, (char*)buffer.value)) == NULL) break;
    if(state->buffer_set(&buffer, "local.soa.%d.mname", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_soa[module->local_in_soa_count].mname = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
    if(state->buffer_set(&buffer, "local.soa.%d.rname", module->local_in_soa_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_soa[module->local_in_soa_count].rname = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
  }

  count = state->config_maxn(state, "local.spf");
  if((module->local_in_spf = (in_txt_t*)calloc((size_t)count, sizeof(module->local_in_spf[0]))) == NULL) CALLOC_FAILED;
  for(module->local_in_spf_count = 0; module->local_in_spf_count < count; module->local_in_spf_count++)
  {
    memset(&module->local_in_spf[module->local_in_spf_count], 0, sizeof(module->local_in_spf[0]));

    if(state->buffer_set(&buffer, "local.spf.%d.ttl", module->local_in_spf_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_spf[module->local_in_spf_count].ttl = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);

    if(state->buffer_set(&buffer, "local.spf.%d.stop", module->local_in_spf_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_spf[module->local_in_spf_count].stop = state->config_get_int(state, (char*)buffer.value, 0);

    if(state->buffer_set(&buffer, "local.spf.%d.pattern", module->local_in_spf_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_spf[module->local_in_spf_count].pattern = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
    for(i = 1; ; i++)
    {
      const char *label;
      uint8_t length;
      if(state->buffer_set(&buffer, "local.spf.%d.txt.%d", module->local_in_spf_count + 1, i) != 0) BUFFER_SET_FAILED;
      if((label = state->config_get_const(state, (char*)buffer.value)) == NULL) break;
      length = (uint8_t)strlen(label);
      if(state->buffer_put_raw(&module->local_in_spf[module->local_in_spf_count].txt, &length, sizeof(length)) != 0) BUFFER_PUT_RAW_FAILED;
      if(state->buffer_put_raw(&module->local_in_spf[module->local_in_spf_count].txt, (const uint8_t*)label, length) != 0) BUFFER_PUT_RAW_FAILED;
    }
  }

  count = state->config_maxn(state, "local.txt");
  if((module->local_in_txt = (in_txt_t*)calloc((size_t)count, sizeof(module->local_in_txt[0]))) == NULL) CALLOC_FAILED;
  for(module->local_in_txt_count = 0; module->local_in_txt_count < count; module->local_in_txt_count++)
  {
    memset(&module->local_in_txt[module->local_in_txt_count], 0, sizeof(module->local_in_txt[0]));

    if(state->buffer_set(&buffer, "local.txt.%d.ttl", module->local_in_txt_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_txt[module->local_in_txt_count].ttl = (uint32_t)state->config_get_int(state, (char*)buffer.value, DEFAULT_TTL);

    if(state->buffer_set(&buffer, "local.txt.%d.stop", module->local_in_txt_count + 1) != 0) BUFFER_SET_FAILED;
    module->local_in_txt[module->local_in_txt_count].stop = state->config_get_int(state, (char*)buffer.value, 0);

    if(state->buffer_set(&buffer, "local.txt.%d.pattern", module->local_in_txt_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->local_in_txt[module->local_in_txt_count].pattern = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
    for(i = 1; ; i++)
    {
      const char *label;
      uint8_t length;
      if(state->buffer_set(&buffer, "local.txt.%d.txt.%d", module->local_in_txt_count + 1, i) != 0) BUFFER_SET_FAILED;
      if((label = state->config_get_const(state, (char*)buffer.value)) == NULL) break;
      length = (uint8_t)strlen(label);
      if(state->buffer_put_raw(&module->local_in_txt[module->local_in_txt_count].txt, &length, sizeof(length)) != 0) BUFFER_PUT_RAW_FAILED;
      if(state->buffer_put_raw(&module->local_in_txt[module->local_in_txt_count].txt, (const uint8_t*)label, length) != 0) BUFFER_PUT_RAW_FAILED;
    }
  }

  state->buffer_free(&buffer);
  state->config_free_node(state, "local");
  return 0;
}

static answer_t answer(int module_index, client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int __attribute__((unused)) lazy)
{
  int i;
  answer_t ret = ANSWER_UNSATISFIED;
  rr_t* rr = NULL;
  state_t *state = client->state;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  if(dclass != CLASS_IN && dclass != CLASS_ANY) return ANSWER_UNSATISFIED;

  if(type == TYPE_A || type == TYPE_ANY)
  {
    for(i = 0; i < module->local_in_a_count; i++)
    {
//jellofish
ERROR_IF(module->trace) (void)state->lprintf(TRACE, "module->local_in_a[%s].stop = %d\n", module->local_in_a[i].pattern, module->local_in_a[i].stop);

      if(state->match(module->local_in_a[i].pattern, (char*)name->value) != 0)
      {
        rr = state->rr_create(name, TYPE_A, CLASS_IN, module->local_in_a[i].ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
        if(rr == NULL) break;
        memcpy(&rr->p.address, &module->local_in_a[i].address, sizeof(rr->p.address));
        if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
        state->free_rr(&rr);
        if(module->local_in_a[i].stop != 0) break;
      }
    }
    if(type != TYPE_ANY)
    {
      if(rr_section == RR_AN && ret == ANSWER_SATISFIED) client->header.flags.aa = 1;
      return ret;
    }
  }
  if(type == TYPE_AAAA || type == TYPE_ANY)
  {
    for(i = 0; i < module->local_in_aaaa_count; i++)
    {
      if(state->match(module->local_in_aaaa[i].pattern, (char*)name->value) != 0)
      {
        rr = state->rr_create(name, TYPE_AAAA, CLASS_IN, module->local_in_aaaa[i].ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
        if(rr == NULL) break;
        memcpy(&rr->p.address, &module->local_in_aaaa[i].address, sizeof(rr->p.address));
        if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
        state->free_rr(&rr);
        if(module->local_in_aaaa[i].stop != 0) break;
      }
    }
    if(type != TYPE_ANY)
    {
      if(rr_section == RR_AN && ret == ANSWER_SATISFIED) client->header.flags.aa = 1;
      return ret;
    }
  }
  if(type == TYPE_CNAME || type == TYPE_ANY)
  {
    for(i = 0; i < module->local_in_cname_count; i++)
    {
      if(state->match(module->local_in_cname[i].pattern, (char*)name->value) != 0)
      {
        rr = state->rr_create(name, TYPE_CNAME, CLASS_IN, module->local_in_cname[i].ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
        if(rr == NULL) break;
        if(state->buffer_set(&rr->p.cname.name, "%s", module->local_in_cname[i].label) == 0)
        {
          if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
        }
        state->free_rr(&rr);
        if(module->local_in_cname[i].stop != 0) break;
      }
    }
    if(type != TYPE_ANY)
    {
      if(rr_section == RR_AN && ret == ANSWER_SATISFIED) client->header.flags.aa = 1;
      return ret;
    }
  }
  if(type == TYPE_MX || type == TYPE_ANY)
  {
    for(i = 0; i < module->local_in_mx_count; i++)
    {
      if(state->match(module->local_in_mx[i].pattern, (char*)name->value) != 0)
      {
        rr = state->rr_create(name, TYPE_MX, CLASS_IN, module->local_in_mx[i].ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
        if(rr == NULL) break;
        rr->p.mx.preference = module->local_in_mx[i].preference;
        if(state->buffer_set(&rr->p.mx.exchange.name, "%s", module->local_in_mx[i].exchange) == 0)
        {
          if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
        }
        state->free_rr(&rr);
        if(module->local_in_mx[i].stop != 0) break;
      }
    }
    if(type != TYPE_ANY)
    {
      if(rr_section == RR_AN && ret == ANSWER_SATISFIED) client->header.flags.aa = 1;
      return ret;
    }
  }
  if(type == TYPE_NS || type == TYPE_ANY)
  {
    for(i = 0; i < module->local_in_ns_count; i++)
    {
      if(state->match(module->local_in_ns[i].pattern, (char*)name->value) != 0)
      {
        rr = state->rr_create(name, TYPE_NS, CLASS_IN, module->local_in_ns[i].ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
        if(rr == NULL) break;
        if(state->buffer_set(&rr->p.ns.name, "%s", module->local_in_ns[i].label) == 0)
        {
          if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
        }
        state->free_rr(&rr);
        if(module->local_in_ns[i].stop != 0) break;
      }
    }
    if(type != TYPE_ANY)
    {
      if(rr_section == RR_AN && ret == ANSWER_SATISFIED) client->header.flags.aa = 1;
      return ret;
    }
  }
  if(type == TYPE_PTR || type == TYPE_ANY)
  {
    for(i = 0; i < module->local_in_ptr_count; i++)
    {
      if(state->match(module->local_in_ptr[i].pattern, (char*)name->value) != 0)
      {
        rr = state->rr_create(name, TYPE_PTR, CLASS_IN, module->local_in_ptr[i].ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
        if(rr == NULL) break;
        if(state->buffer_set(&rr->p.ptr.name, "%s", module->local_in_ptr[i].label) == 0)
        {
          if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
        }
        state->free_rr(&rr);
        if(module->local_in_ptr[i].stop != 0) break;
      }
    }
    if(type != TYPE_ANY)
    {
      if(rr_section == RR_AN && ret == ANSWER_SATISFIED) client->header.flags.aa = 1;
      return ret;
    }
  }
  if(type == TYPE_SOA || type == TYPE_ANY)
  {
    for(i = 0; i < module->local_in_soa_count; i++)
    {
      if(state->match(module->local_in_soa[i].pattern, (char*)name->value) != 0)
      {
        rr = state->rr_create(name, TYPE_SOA, CLASS_IN, module->local_in_soa[i].ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
        if(rr == NULL) break;
        if(state->buffer_set(&rr->p.soa.mname.name, "%s", module->local_in_soa[i].mname) != 0 || state->buffer_set(&rr->p.soa.rname.name, "%s", module->local_in_soa[i].rname) == 0)
        {
          rr->p.soa.serial = module->local_in_soa[i].serial;
          rr->p.soa.refresh = module->local_in_soa[i].refresh;
          rr->p.soa.retry = module->local_in_soa[i].retry;
          rr->p.soa.expire = module->local_in_soa[i].expire;
          rr->p.soa.minimum = module->local_in_soa[i].minimum;
          if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
        }
        state->free_rr(&rr);
        if(module->local_in_soa[i].stop != 0) break;
      }
    }
    if(type != TYPE_ANY)
    {
      if(rr_section == RR_AN && ret == ANSWER_SATISFIED) client->header.flags.aa = 1;
      return ret;
    }
  }
  if(type == TYPE_SPF || type == TYPE_ANY)
  {
    for(i = 0; i < module->local_in_spf_count; i++)
    {
      if(state->match(module->local_in_spf[i].pattern, (char*)name->value) != 0)
      {
        rr = state->rr_create(name, TYPE_SPF, CLASS_IN, module->local_in_spf[i].ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
        if(rr == NULL) break;
        if(state->buffer_copy(&rr->p.spf, &module->local_in_spf[i].txt) == 0)
        {
          if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
        }
        state->free_rr(&rr);
        if(module->local_in_spf[i].stop != 0) break;
      }
    }
    if(type != TYPE_ANY)
    {
      if(rr_section == RR_AN && ret == ANSWER_SATISFIED) client->header.flags.aa = 1;
      return ret;
    }
  }
  if(type == TYPE_TXT || type == TYPE_ANY)
  {
    for(i = 0; i < module->local_in_txt_count; i++)
    {
      if(state->match(module->local_in_txt[i].pattern, (char*)name->value) != 0)
      {
        rr = state->rr_create(name, TYPE_TXT, CLASS_IN, module->local_in_txt[i].ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
        if(rr == NULL) break;
        if(state->buffer_copy(&rr->p.txt, &module->local_in_txt[i].txt) == 0)
        {
          if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
        }
        state->free_rr(&rr);
        if(module->local_in_txt[i].stop != 0) break;
      }
    }
  }
  if(rr_section == RR_AN && ret == ANSWER_SATISFIED) client->header.flags.aa = 1;
  return ret;
}

module_t unnamed_module = { "local", NULL, init, answer, NULL, term };
